use std::collections::VecDeque;

use anyhow::Context;
use subprocess::{Communicator, Popen, PopenConfig, Redirection};

pub struct RtlSdr {
    process: Popen,
    comms: Communicator,
    out_buf: VecDeque<u8>,
    err_buf: Vec<u8>,
}

impl RtlSdr {
    pub fn new(
        freq: u32,
        device_id: &str,
        gain: Option<f32>,
        bias_t: bool,
    ) -> anyhow::Result<Self> {
        let freq = format!("{freq}");
        let mut args = vec![
            "rtl_fm", "-f", &freq, "-M", "raw", "-F9", "-p", "0", "-d", device_id, "-s", "48000",
        ];
        let mut gain_s = String::new();
        if let Some(g) = gain {
            args.push("-g");
            gain_s.push_str(&format!("{:.2}", g));
            args.push(&gain_s);
        }
        if bias_t {
            args.push("-T");
        }
        let mut process = Popen::create(
            &args,
            PopenConfig {
                stdout: Redirection::Pipe,
                stderr: Redirection::Pipe,
                detached: true,
                ..Default::default()
            },
        )
        .context("Error starting rtl-fm")?;

        let comms = process.communicate_start(None).limit_size(1024);

        Ok(Self {
            process,
            comms,
            out_buf: VecDeque::new(),
            err_buf: vec![],
        })
    }
}

impl Iterator for RtlSdr {
    type Item = u8;

    fn next(&mut self) -> Option<Self::Item> {
        while self.out_buf.is_empty() {
            let (out, err) = self.comms.read().ok()?;
            let out = out.expect("Missing stdout");
            let err = err.expect("Missing stderr");

            self.out_buf.extend(out.iter());
            for e in err {
                if e == b'\n' {
                    handle_err(&self.err_buf);
                    self.err_buf.clear();
                } else {
                    self.err_buf.push(e);
                }
            }

            if self.process.poll().is_some() {
                handle_err(&self.err_buf);
                self.err_buf.clear();
                return None;
            }
        }

        Some(self.out_buf.pop_front().unwrap())
    }
}

fn handle_err(err: &[u8]) {
    if !err.is_empty() {
        eprintln!("[rtl_fm] {}", String::from_utf8_lossy(err));
    }
}
